from json import JSONEncoder
from typing import Any, List, Optional

from .clause_tokenizer import ClauseTokenizer
from .utils import CaseInsensitiveDict, Variable

ACCEPTED_QUOTES = ["'", '"', "’"]
NEW_LINE = "\n"


class JsonParser:
    """
    Input can be a multiple documents, separated by a new-line (\n) characters
    So we can't use the builtin JSON parser
    """

    def parse(
        self,
        original: str,
        tokenizer: Optional[ClauseTokenizer] = None,
        only_parse_initial: bool = False,
    ) -> Any:
        if not (original.startswith("{") or original.startswith("[")):
            # Doesn't look like JSON - let's return as a variable
            return original if original.isnumeric() else Variable(original)
        section: Optional[str] = None  # DICT_KEY | KEY_TO_VALUE | DICT_VAL | OBJECT_END
        dict_key = ""
        current_phrase = ""
        result = CaseInsensitiveDict()
        tokenizer = tokenizer or ClauseTokenizer(original)
        while True:
            c = tokenizer.next()
            if not c:
                break
            elif c == "[" and (not section or section == "KEY_TO_VALUE"):
                # Start of a list
                if not section:
                    return self._parse_list(original, tokenizer)
                else:
                    result[dict_key] = self._parse_list(original, tokenizer)
                    section = None
                    current_phrase = ""
            elif c in ["{", ","] and (not section or section == "OBJECT_END"):
                # Start of a key
                section = "DICT_KEY"
                tokenizer.skip_until(ACCEPTED_QUOTES)
                current_phrase = ""
            elif c in ACCEPTED_QUOTES and section == "DICT_KEY":
                # End of a key
                dict_key = current_phrase
                tokenizer.skip_until([":"])
                section = "KEY_TO_VALUE"
                current_phrase = ""
            elif c in ["{"] and section == "KEY_TO_VALUE":
                # Start of a value with a new dictionary
                tokenizer.revert()  # Ensure we start the new parser with the initial {
                result[dict_key] = self.parse(original, tokenizer)
                section = None
                current_phrase = ""
            elif c in ACCEPTED_QUOTES and section == "KEY_TO_VALUE":
                # Start of a value
                section = "DICT_VAL"
                current_phrase = ""
            elif c in ACCEPTED_QUOTES and section == "DICT_VAL":
                # End of a value
                result[dict_key] = current_phrase
                section = None
                current_phrase = ""
            elif c in ["}"] and section in ["VAR_VALUE", "INT_VALUE"]:
                # End of a variable/number
                if section == "INT_VALUE":
                    result[dict_key] = int(current_phrase)
                elif current_phrase.lower() in ["true", "false"]:
                    result[dict_key] = current_phrase.lower() == "true"
                else:
                    result[dict_key] = Variable(current_phrase)
                section = None
                current_phrase = ""
                if only_parse_initial:
                    break
                else:
                    tokenizer.revert()
            elif c in [","] and section in ["VAR_VALUE", "INT_VALUE"]:
                if section == "INT_VALUE":
                    result[dict_key] = int(current_phrase)
                elif current_phrase.lower() in ["true", "false"]:
                    result[dict_key] = current_phrase.lower() == "true"
                else:
                    result[dict_key] = Variable(current_phrase)
                tokenizer.revert()
                section = None
                current_phrase = ""
            elif section in ["OBJECT_END"]:
                next_documents = self.parse(original, tokenizer)
                if next_documents == {}:
                    return result
                elif isinstance(next_documents, list):
                    return [result] + next_documents
                else:
                    return [result, next_documents]
            elif c == "}" and section is None:
                section = "OBJECT_END"
                # We know whether we are at the end of an object at this point
                # But we don't know whether this is:
                # - end of the root object
                # - end of a nested object
                # - inbetween multiple objects (separated by new-line)
                tokenizer.skip_white_space()
                if tokenizer.current() == "{":
                    # we're inbetween multiple objects - continue parsing
                    tokenizer.revert()
                    pass
                else:
                    # we're at the end of the root object - next char is probably None. Break and return to the user
                    # we're at the end of a nested object - next char is probably }.    Break and let the parent processor takeover
                    break
            elif c in [" ", NEW_LINE] and section not in ["DICT_KEY", "DICT_VAL"]:
                pass
            else:
                if section == "KEY_TO_VALUE":
                    # We found a value directly after the key, unquoted
                    # That means it's either a variable, or a number
                    if c.isnumeric():
                        section = "INT_VALUE"
                    else:
                        section = "VAR_VALUE"
                if section in ["DICT_KEY", "DICT_VAL", "INT_VALUE", "VAR_VALUE"]:
                    current_phrase += c

        return result

    def _parse_list(self, original: str, tokenizer: ClauseTokenizer) -> Any:
        result: List[Any] = list()
        section = None
        current_phrase = ""
        while True:
            c = tokenizer.next()
            if not c:
                break
            if c == "{":
                tokenizer.revert()  # Ensure we start the new parser with the initial {
                result.append(self.parse(original, tokenizer, only_parse_initial=True))
                if tokenizer.current() == "]":
                    break
                tokenizer.skip_until([","])
                tokenizer.skip_white_space()
            elif c in ACCEPTED_QUOTES and section is None:
                section = "VALUE"
            elif c in ACCEPTED_QUOTES and section == "VALUE":
                result.append(current_phrase)
                current_phrase = ""
                section = None
            elif c == "]" and not section:
                return result
            elif c == "]" and section == "VAR_VALUE":
                if current_phrase.isnumeric():
                    result.append(int(current_phrase))
                elif current_phrase.lower() in ["true", "false"]:
                    result.append(current_phrase.lower() == "true")
                else:
                    result.append(Variable(current_phrase))
                return result
            elif c == "," and section == "VAR_VALUE":
                if current_phrase.isnumeric():
                    result.append(int(current_phrase))
                elif current_phrase.lower() in ["true", "false"]:
                    result.append(current_phrase.lower() == "true")
                else:
                    result.append(Variable(current_phrase))
                current_phrase = ""
                section = None
                tokenizer.skip_white_space()
            elif c == "," and not section:
                tokenizer.skip_white_space()
            elif not section:
                current_phrase += c
                section = "VAR_VALUE"
            elif section in ["VALUE", "VAR_VALUE"]:
                current_phrase += c
        return result


class SelectEncoder(JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, Variable) and o.value is None:
            return None
        if isinstance(o, CaseInsensitiveDict):
            return dict(o.items())
        return JSONEncoder.default(self, o)



import re
import sys
from collections import OrderedDict
from copy import copy
from typing import Any, Dict, List, Iterator, Optional, Tuple, Union, TYPE_CHECKING

if sys.version_info[:2] > (3, 8):
    from collections.abc import Mapping, MutableMapping
else:
    from typing import Mapping, MutableMapping

if TYPE_CHECKING:
    from .where_parser import AbstractWhereClause

from .._packages.boto3.types import TypeDeserializer, TypeSerializer


deserializer = TypeDeserializer()
serializer = TypeSerializer()


def is_dict(dct: Any) -> bool:
    return isinstance(dct, dict) or isinstance(dct, CaseInsensitiveDict)


def find_nested_data(
    select_clause: str, data_source: List["CaseInsensitiveDict"]
) -> List[Dict[str, Any]]:
    """
    Iterate over a list of JSON objects, and return only the keys specified
    :param select_clause: Key of the data source, in dot-notation: a.b
    :param data_source: List of JSON documents as dictionary
    :return: A list of JSON keys as dictionary
    """

    results: List[Dict[str, Any]] = []

    for row in data_source:
        # Run the select-clause over each row
        result = find_nested_data_in_object(select_clause=select_clause, json_doc=row)
        results.append(result)
    return results


def find_nested_data_in_object(
    select_clause: str, json_doc: Union[str, "CaseInsensitiveDict"]
) -> Any:
    if select_clause == "*":
        return json_doc

    current_key = select_clause.split(".")[0]
    remaining_keys = ".".join(select_clause.split(".")[1:])
    if isinstance(json_doc, list):  # type: ignore[unreachable]
        result = []  # type: ignore[unreachable]
        for row in json_doc:
            if current_key not in row:
                result.append(MissingVariable())
            else:
                result.append(
                    find_nested_data_in_object(
                        json_doc=row[current_key], select_clause=remaining_keys
                    )
                )
        return result
    elif is_dict(json_doc):
        if current_key not in json_doc:
            return MissingVariable()
        if remaining_keys:
            return find_nested_data_in_object(
                json_doc=json_doc[current_key], select_clause=remaining_keys  # type: ignore[index]
            )
        return json_doc.get_original(current_key)  # type: ignore[union-attr]


def find_value_in_document(keys: List[str], json_doc: Any) -> Any:
    if not is_dict(json_doc):
        return None
    key_is_array = re.search(r"(.+)\[(\d+)\]$", keys[0])
    if key_is_array:
        key_name = key_is_array.group(1)
        array_index = int(key_is_array.group(2))
        try:
            requested_list = json_doc.get(key_name, [])
            assert isinstance(requested_list, list)
            doc_one_layer_down = requested_list[array_index]
        except IndexError:
            # Array exists, but does not have enough values
            doc_one_layer_down = {}
        except AssertionError:
            # Requested key is not an array - fail silently just like AWS does
            doc_one_layer_down = {}
        return find_value_in_document(keys[1:], doc_one_layer_down)
    if len(keys) == 1:
        return json_doc.get(keys[0])
    return find_value_in_document(keys[1:], json_doc.get(keys[0], {}))


def find_value_in_dynamodb_document(keys: List[str], json_doc: Any) -> Any:
    if not is_dict(json_doc):
        return None
    key_is_array = re.search(r"(.+)\[(\d+)\]$", keys[0])
    if key_is_array:
        key_name = key_is_array.group(1)
        array_index = int(key_is_array.group(2))
        try:
            requested_list = json_doc.get(key_name, {})
            assert "L" in requested_list
            doc_one_layer_down = requested_list["L"][array_index]
            if "M" in doc_one_layer_down:
                doc_one_layer_down = doc_one_layer_down["M"]
        except IndexError:
            # Array exists, but does not have enough values
            doc_one_layer_down = {}
        except AssertionError:
            # Requested key is not a list - fail silently just like AWS does
            doc_one_layer_down = {}
        return find_value_in_dynamodb_document(keys[1:], doc_one_layer_down)
    if len(keys) == 1:
        if "M" in json_doc:
            return json_doc["M"].get(keys[0])
        else:
            return json_doc.get(keys[0])
    nested_doc = json_doc.get(keys[0], {})
    if "M" in nested_doc:
        return find_value_in_dynamodb_document(keys[1:], nested_doc["M"])
    # Key is not a map
    # Or does not exist
    return None


class QueryMetadata:
    def __init__(
        self,
        tables: Dict[str, str],
        where_clause: Optional["AbstractWhereClause"] = None,
        is_select_query: bool = False,
    ):
        self._tables = tables
        self._where_clause = where_clause
        self._is_select_query = is_select_query

    def get_table_names(self) -> List[str]:
        return list(self._tables.values())

    def get_filter_names(self) -> List[str]:
        if self._where_clause:
            return self._where_clause.get_filter_names()
        return []

    def is_select_query(self) -> bool:
        return self._is_select_query


class Variable:
    def __init__(self, value: Any) -> None:
        self.value = value
        if value == "null":
            self.value = None
        elif isinstance(value, str) and value.lower() in ["true", "false"]:
            self.value = bool(value)

    def __repr__(self) -> str:
        return f"<{self.value}>"

    def __hash__(self) -> int:
        return hash(self.value)

    def __eq__(self, other: Any) -> bool:
        return other and isinstance(other, Variable) and self.value == other.value

    def apply(self, value: Any) -> Any:
        if isinstance(value, dict):
            split_value = (
                self.value.split(".") if isinstance(self.value, str) else [self.value]
            )
            current_key = split_value[0]
            if current_key not in value:
                return MissingVariable()
            remaining_keys = ".".join(split_value[1:])
            return Variable(remaining_keys).apply(value[current_key])
        else:
            return value


class MissingVariable(Variable):
    def __init__(self) -> None:
        super().__init__(value=None)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, MissingVariable)


class CaseInsensitiveDict(MutableMapping[str, Any]):
    # Taken from https://raw.githubusercontent.com/kennethreitz/requests/v2.25.1/requests/structures.py

    def __init__(self, data: Optional[Mapping[str, Any]] = None):
        self._store: Dict[str, Any] = OrderedDict()
        if data:
            self.update(
                {
                    key: CaseInsensitiveDict(val) if is_dict(val) else val
                    for key, val in data.items()
                }
            )

    def __setitem__(self, key: str, value: Any) -> None:
        # Use the lowercased key for lookups, but store the actual
        # key alongside the value.
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key: str) -> Any:
        return self._store[key.lower()][1]

    def __delitem__(self, key: str) -> None:
        del self._store[key.lower()]

    def __iter__(self) -> Iterator[str]:
        return (casedkey for casedkey, mappedvalue in self._store.values())

    def __len__(self) -> int:
        return len(self._store)

    def lower_items(self) -> Iterator[Tuple[str, Any]]:
        """Like iteritems(), but with all lowercase keys."""
        return ((lowerkey, keyval[1]) for (lowerkey, keyval) in self._store.items())

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    def get_regular(self) -> Dict[str, Any]:
        dct = {}
        for key, val in self.items():
            if isinstance(val, CaseInsensitiveDict):
                dct[key] = val.get_regular()
            else:
                dct[key] = val
        return dct

    def get_original(self, key: str) -> "CaseInsensitiveDict":
        original_key, original_value = self._store[key.lower()]
        return CaseInsensitiveDict({original_key: original_value})

    # Copy is required
    def copy(self) -> "CaseInsensitiveDict":
        _new = CaseInsensitiveDict()
        for key, value in self._store.values():
            _new[key] = copy(value)
        return _new

    def __repr__(self) -> str:
        return str(dict(self.items()))



import re

from typing import Dict, Any, List, Optional, Tuple

from ..exceptions import ParserException
from .delete_parser import DeleteParser
from .from_parser import DynamoDBFromParser, S3FromParser, FromParser
from .insert_parser import InsertParser
from .select_parser import SelectParser
from .update_parser import UpdateParser
from .where_parser import DynamoDBWhereParser, S3WhereParser, WhereParser
from .utils import is_dict, QueryMetadata, CaseInsensitiveDict


TYPE_RESPONSE = Tuple[
    List[Dict[str, Any]],
    Dict[str, List[Tuple[Optional[Dict[str, Any]], Optional[Dict[str, Any]]]]],
]


class S3SelectParser:
    def __init__(self, source_data: Dict[str, str]):
        # Source data is in the format: {source: json}
        # Where 'json' is one or more json documents separated by a newline
        self.documents = source_data
        self.table_prefix = "s3object"

    def parse(self, query: str) -> List[Dict[str, Any]]:
        query = query.replace("\n", " ")
        clauses = re.split("SELECT | FROM | WHERE ", query, flags=re.IGNORECASE)
        # First clause is whatever comes in front of SELECT - which should be nothing
        _ = clauses[0]
        # FROM
        from_parser = S3FromParser(from_clause=clauses[2])

        source_data = from_parser.get_source_data(self.documents)
        if is_dict(source_data):
            source_data = [source_data]

        # WHERE
        if len(clauses) > 3:
            where_clause = clauses[3]
            source_data = S3WhereParser(source_data).parse(where_clause)

        # SELECT
        select_clause = clauses[1]
        table_prefix = self.table_prefix
        for alias_key, alias_value in from_parser.clauses.items():
            if table_prefix == alias_value:
                table_prefix = alias_key
        return SelectParser(table_prefix).parse(
            select_clause, from_parser.clauses, source_data
        )


class DynamoDBStatementParser:
    def __init__(self, source_data: Dict[str, List[Dict[str, Any]]]):
        """
        Source Data should be a list of DynamoDB documents, mapped to the table name
        {
          "table_name": [
             {
               "hash_key": "..",
               "other_item": {"S": ".."},
               ..
             },
             ..
          ],
          ..
        }
        """
        # Source data is in the format: {source: json}
        # Where 'json' is one or more json documents separated by a newline
        self.documents = {
            key: [CaseInsensitiveDict(v) for v in val]
            for key, val in source_data.items()
        }

    def parse(  # type: ignore[return]
        self, query: str, parameters: Optional[List[Dict[str, Any]]] = None
    ) -> TYPE_RESPONSE:
        if query.lower().startswith("select"):
            return_data, updates = self._parse_select(query, parameters)
            for item in return_data:
                for key, val in item.items():
                    item[key] = val.get_regular()
            return return_data, updates

        if query.lower().startswith("update"):
            return self._parse_update(query, parameters)

        if query.lower().startswith("delete"):
            return self._parse_delete(query)

        if query.lower().startswith("insert"):
            return self._parse_insert(query)

    def _parse_select(
        self, query: str, parameters: Optional[List[Dict[str, Any]]] = None
    ) -> TYPE_RESPONSE:
        query = query.replace("\n", " ")
        clauses = re.split("SELECT | FROM | WHERE ", query, flags=re.IGNORECASE)
        # First clause is whatever comes in front of SELECT - which should be nothing
        _ = clauses[0]
        # FROM
        from_parser = DynamoDBFromParser(from_clause=clauses[2])

        source_data = self.documents[list(from_parser.clauses.values())[0].lower()]

        # WHERE
        if len(clauses) > 3:
            where_clause = clauses[3]
            source_data = DynamoDBWhereParser(source_data).parse(
                where_clause, parameters
            )

        # SELECT
        select_clause = clauses[1]
        queried_data = SelectParser().parse(
            select_clause, from_parser.clauses, source_data
        )
        updates: Dict[
            str, List[Tuple[Optional[Dict[str, Any]], Optional[Dict[str, Any]]]]
        ] = {}
        return queried_data, updates

    def _parse_update(
        self, query: str, parameters: Optional[List[Dict[str, Any]]] = None
    ) -> TYPE_RESPONSE:
        query = query.replace("\n", " ")

        table_name, attrs_to_update, attrs_to_filter = UpdateParser().parse(query)

        parameters_requested = len(
            [_ for _, val in attrs_to_update + attrs_to_filter if val == "?"]
        )
        if parameters_requested and len(parameters) != parameters_requested:  # type: ignore
            raise ParserException(
                name="ValidationError",
                message="Number of parameters in request and statement don't match.",
            )

        attrs_to_update = [
            (key, parameters.pop(0) if val == "?" else val)  # type: ignore
            for key, val in attrs_to_update
        ]
        attrs_to_filter = [
            (key, parameters.pop(0) if val == "?" else val)  # type: ignore
            for key, val in attrs_to_filter
        ]

        source_data = self.documents[table_name]
        updates_per_table: Dict[
            str, List[Tuple[Optional[Dict[str, Any]], Optional[Dict[str, Any]]]]
        ] = {table_name: []}

        for item in source_data:
            if all([item.get(name) == val for name, val in attrs_to_filter]):
                new_item = item.copy()
                for attr_key, attr_value in attrs_to_update:
                    if attr_value is None:
                        new_item.pop(attr_key, None)
                    else:
                        new_item[attr_key] = attr_value
                updates_per_table[table_name].append(
                    (item.get_regular(), new_item.get_regular())
                )

        return [], updates_per_table

    def _parse_delete(self, query: str) -> TYPE_RESPONSE:
        query = query.replace("\n", " ")

        table_name, attrs_to_filter = DeleteParser().parse(query)

        source_data = self.documents[table_name]
        deletes_per_table: Dict[
            str, List[Tuple[Optional[Dict[str, Any]], Optional[Dict[str, Any]]]]
        ] = {table_name: []}

        for item in source_data:
            if all([item.get(name) == val for name, val in attrs_to_filter]):
                deletes_per_table[table_name].append((item.get_regular(), None))

        return [], deletes_per_table

    def _parse_insert(self, query: str) -> TYPE_RESPONSE:
        query = query.replace("\n", " ")

        table_name, new_item = InsertParser().parse(query)
        return [], {table_name: [(None, new_item)]}

    @classmethod
    def get_query_metadata(cls, query: str) -> QueryMetadata:
        query = query.replace("\n", " ")
        if query.lower().startswith("select"):
            clauses = re.split("SELECT | FROM | WHERE ", query, flags=re.IGNORECASE)

            from_parser = FromParser(clauses[2])
            # WHERE
            if len(clauses) > 3:
                where_clause = clauses[3]
                where = WhereParser.parse_where_clause(where_clause)
            else:
                where = None

            return QueryMetadata(
                tables=from_parser.clauses, where_clause=where, is_select_query=True
            )
        elif query.lower().startswith("update"):
            table_name, attrs_to_update, attrs_to_filter = UpdateParser().parse(query)
            return QueryMetadata(tables={table_name: table_name}, where_clause=None)
        elif query.lower().startswith("delete"):
            query = query.replace("\n", " ")

            table_name, attrs_to_filter = DeleteParser().parse(query)
            return QueryMetadata(tables={table_name: table_name})
        elif query.lower().startswith("insert"):
            query = query.replace("\n", " ")

            table_name, new_item = InsertParser().parse(query)
            return QueryMetadata(tables={table_name: table_name})
        raise Exception



from typing import Dict, Any, List, Optional

from .clause_tokenizer import ClauseTokenizer
from .utils import (
    find_nested_data_in_object,
    MissingVariable,
    is_dict,
    CaseInsensitiveDict,
)


class SelectClause:
    def __init__(self, value: str, table_prefix: Optional[str] = None):
        self.table_prefix = table_prefix
        self.value = value.strip()

    def select(self, document: CaseInsensitiveDict) -> Any:
        if self.value == "*":
            if self.table_prefix and self.table_prefix in document:
                return document[self.table_prefix]
            else:
                return document
        if "." in self.value:
            key, remaining = self.value.split(".", maxsplit=1)
            return find_nested_data_in_object(
                select_clause=remaining, json_doc=document[key]
            )
        elif not self.table_prefix:
            return find_nested_data_in_object(
                select_clause=self.value, json_doc=document
            )
        else:
            if is_dict(document[self.value]):
                return document[self.value]
            else:
                return {self.value: document[self.value]}

    def __repr__(self) -> str:
        return f"<SelectClause({self.value})>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, SelectClause) and other.value == self.value


class FunctionClause(SelectClause):
    def __init__(self, value: str, function_name: str):
        super().__init__(value)
        self.function_name = function_name.strip()

    def execute(
        self, aliases: Dict[str, str], documents: List[CaseInsensitiveDict]
    ) -> Dict[str, int]:
        return {"_1": len(documents)}

    def __repr__(self) -> str:
        return f"<FunctionClause({self.function_name}({self.value}))>"

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, FunctionClause)
            and other.value == self.value
            and other.function_name == self.function_name
        )


class SelectParser:
    def __init__(self, table_prefix: Optional[str] = None):
        self.table_prefix = table_prefix

    def parse(
        self,
        select_clause: str,
        aliases: Dict[str, Any],
        documents: List[CaseInsensitiveDict],
    ) -> List[Dict[str, Any]]:
        clauses = SelectParser.parse_clauses(select_clause, prefix=self.table_prefix)

        for clause in clauses:
            if isinstance(clause, FunctionClause):
                return [clause.execute(aliases, documents)]

        result: List[Dict[str, Any]] = []

        for json_document in documents:
            filtered_document = dict()
            for clause in clauses:
                attr = clause.select(json_document)
                if attr is not None and not isinstance(attr, MissingVariable):
                    # Specific usecase:
                    # select * from s3object[*].Name my_n
                    if (
                        "." in list(aliases.values())[0]
                        and list(aliases.keys())[0] in attr
                        and select_clause == "*"
                    ):
                        filtered_document.update({"_1": attr[list(aliases.keys())[0]]})
                    else:
                        filtered_document.update(attr)
            result.append(filtered_document)
        return result

    @classmethod
    def parse_clauses(
        cls, select_clause: str, prefix: Optional[str] = None
    ) -> List[SelectClause]:
        results = []
        tokenizer = ClauseTokenizer(select_clause)
        current_clause = fn_name = ""
        while True:
            c = tokenizer.next()
            if not c or c in [","]:
                if current_clause != "":
                    results.append(
                        SelectClause(value=current_clause, table_prefix=prefix)
                    )
                if not c:
                    break
                else:
                    current_clause = ""
                    continue
            if c in ["("]:
                fn_name = current_clause
                current_clause = ""
                continue
            if c in [")"]:
                results.append(
                    FunctionClause(function_name=fn_name, value=current_clause)
                )
                current_clause = ""
                continue
            current_clause += c
        return results



from typing import Any, Dict, List, Tuple

from .clause_tokenizer import ClauseTokenizer
from .utils import serializer


class DeleteParser:
    def parse(self, query: str) -> Tuple[str, List[Tuple[str, Dict[str, Any]]]]:
        tokenizer = ClauseTokenizer(query)

        section = "START"
        current_phrase = ""

        table_name = ""
        attr_filters: List[Tuple[str, Dict[str, Any]]] = []

        while True:
            c = tokenizer.next()
            if c is None:
                break

            if c == " ":
                if section == "START":
                    assert (
                        current_phrase.upper() == "DELETE"
                    ), f"{current_phrase} should be DELETE"
                    current_phrase = ""
                    section = "DELETE_FROM"
                    continue
                if section == "DELETE_FROM":
                    assert (
                        current_phrase.upper() == "FROM"
                    ), f"{current_phrase} should be FROM"
                    current_phrase = ""
                    section = "TABLE_NAME"
                    continue
                if section == "ACTION":
                    assert current_phrase.upper() in ["SET", "REMOVE"]
                    section = f"ACTION_{current_phrase.upper()}"
                    current_phrase = ""
                    continue
                if section == "SECTION_WHERE":
                    assert current_phrase.upper() == "WHERE"
                    section = "WHERE"
                    current_phrase = ""
                    continue
                if section == "WHERE_AND":
                    assert current_phrase.upper() == "AND"
                    section = "WHERE"
                    current_phrase = ""
                if section == "TABLE_NAME":
                    table_name = current_phrase
                    current_phrase = ""
                    tokenizer.skip_white_space()
                    section = "SECTION_WHERE"
                continue
            elif c in ["'", '"']:
                if section == "TABLE_NAME":
                    table_name = tokenizer.next_until([c])
                    tokenizer.skip_white_space()
                    section = "SECTION_WHERE"
                continue
            elif c == "=":
                if section == "WHERE":
                    attr_name = current_phrase
                    tokenizer.skip_white_space()
                    quote_type = tokenizer.current()
                    assert quote_type in ["'", '"']

                    tokenizer.next()
                    attr_value = tokenizer.next_until([quote_type])
                    attr_filters.append((attr_name, serializer.serialize(attr_value)))

                    tokenizer.skip_white_space()
                    current_phrase = ""
                    section = "WHERE_AND"
                continue
            else:
                current_phrase += c
                continue

        return table_name, attr_filters



import json
from typing import Any, Dict, List


def csv_to_json(input: str, headers_included: bool = False) -> str:
    output = ""
    headers: List[str] = []
    for line in input.split("\n"):
        values = line.split(",")
        if len(headers) == 0:
            if headers_included:
                headers = values
                continue
            else:
                headers = [f"_{x}" for x in range(1, len(values) + 1)]
        line = json.dumps({f"{headers[idx]}": key for idx, key in enumerate(values)})
        output += f"{line}\n"
    if output.endswith("\n"):
        output = output.rstrip("\n")
    return output


def json_to_csv(
    input: List[Dict[str, Any]], field_delimiter: str, record_delimiter: str
) -> str:
    result = ""
    for row in input:
        result += (
            field_delimiter.join([f"{v}" for v in row.values()]) + record_delimiter
        )
    return result



from typing import Any, Dict, Tuple

from .json_parser import JsonParser
from .clause_tokenizer import ClauseTokenizer
from .utils import serializer


class InsertParser:
    def parse(self, query: str) -> Tuple[str, Dict[str, Any]]:
        tokenizer = ClauseTokenizer(query)

        section = "START"
        current_phrase = ""

        table_name = ""
        attr = {}

        while True:
            c = tokenizer.next()
            if c is None:
                break

            if c == " ":
                if section == "START":
                    assert (
                        current_phrase.upper() == "INSERT"
                    ), f"{current_phrase} should be INSERT"
                    current_phrase = ""
                    section = "INSERT_FROM"
                    continue
                if section == "INSERT_FROM":
                    assert (
                        current_phrase.upper() == "INTO"
                    ), f"{current_phrase} should be INTO"
                    current_phrase = ""
                    section = "TABLE_NAME"
                    continue
                if section == "SECTION_VALUE":
                    assert current_phrase.upper() in ["VALUE"]
                    tokenizer.skip_white_space()
                    attr = JsonParser().parse(tokenizer.give_remaining())
                    for key, value in attr.items():
                        attr[key] = serializer.serialize(value)
                if section == "TABLE_NAME":
                    table_name = current_phrase
                    current_phrase = ""
                    tokenizer.skip_white_space()
                    section = "SECTION_VALUE"
                continue
            elif c in ["'", '"']:
                if section == "TABLE_NAME":
                    table_name = tokenizer.next_until([c])
                    tokenizer.skip_white_space()
                    section = "SECTION_VALUE"
                continue
            else:
                current_phrase += c
                continue

        return table_name, attr



from typing import Optional, List


class ClauseTokenizer:
    def __init__(self, from_clause: str):
        self.token_list = from_clause
        self.token_pos = 0

    def current(self) -> Optional[str]:
        """
        Returns the current char - or None
        """
        try:
            return self.token_list[self.token_pos]
        except IndexError:
            return None

    def next(self) -> Optional[str]:
        """
        Returns the next token - or None
        :return:
        """
        try:
            crnt_token = self.token_list[self.token_pos]
            self.token_pos += 1
            return crnt_token
        except IndexError:
            return None

    def peek(self) -> Optional[str]:
        try:
            return self.token_list[self.token_pos + 1]
        except IndexError:
            return None

    def revert(self) -> None:
        self.token_pos -= 1

    def skip_white_space(self) -> None:
        try:
            while self.token_list[self.token_pos] in [" ", "\n"]:
                self.token_pos += 1
        except IndexError:
            pass

    def give_remaining(self) -> str:
        return self.next_until(chars=[None])

    def next_until(self, chars: List[Optional[str]]) -> str:
        """
        Return the following characters up until (but not including) any of the characters defined in chars
        :param chars:
        :return:
        """
        phrase = ""
        crnt = self.next()
        while crnt and crnt not in chars:
            phrase += crnt
            crnt = self.next()
        return phrase

    def skip_until(self, chars: List[str]) -> None:
        """
        Skip until we encounter one of the provided chars. Inclusive, so we also skip the first char encountered
        :param chars:
        """
        crnt = self.next()
        while crnt and crnt not in chars:
            crnt = self.next()



from decimal import Decimal
from typing import Any, Callable, Dict, List, Optional

from .clause_tokenizer import ClauseTokenizer
from .utils import find_value_in_document
from .utils import find_value_in_dynamodb_document, CaseInsensitiveDict
from .utils import serializer


class AbstractWhereClause:
    def __init__(self) -> None:
        self.children: List[AbstractWhereClause] = []

    def apply(self, find_value: Any, row: Any) -> bool:
        return NotImplemented

    def get_filter_names(self) -> List[str]:
        all_names: List[str] = []
        for child in self.children:
            all_names.extend(child.get_filter_names())
        return all_names

    def process_value(self, fn: Callable[[str], Dict[str, Any]]) -> None:
        """
        Transform all the values in this Where-clause, using a custom function
        """
        for child in self.children:
            child.process_value(fn)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, AbstractWhereClause):
            return self.children == other.children
        return NotImplemented

    def __str__(self) -> str:
        return f"<{type(self)} {self.children}>"

    def __repr__(self) -> str:
        return str(self)


class WhereAndClause(AbstractWhereClause):
    def __init__(self, left: Optional["AbstractWhereClause"]):
        super().__init__()
        self.children.append(left)  # type: ignore[arg-type]

    def apply(self, find_value: Callable[[str], Dict[str, Any]], row: Any) -> bool:
        return all([child.apply(find_value, row) for child in self.children])


class WhereOrClause(AbstractWhereClause):
    def __init__(self, left: Optional["AbstractWhereClause"]):
        super().__init__()
        self.children.append(left)  # type: ignore[arg-type]

    def apply(self, find_value: Any, row: Any) -> bool:
        return any([child.apply(find_value, row) for child in self.children])


class WhereClause(AbstractWhereClause):
    def __init__(self, fn: str, left: List[str], right: Any):
        super().__init__()
        self.fn = fn.lower()
        self.left = left
        self.right = right

    def apply(self, find_value: Any, row: Any) -> bool:
        value = find_value(self.left, row)
        if self.fn == "contains":
            if "S" in self.right and "S" in value:
                return self.right["S"] in value["S"]
        if self.fn == "is":
            if self.right == {"S": "MISSING"}:
                return value is None
            elif self.right == {"S": "NOT MISSING"}:
                return value is not None
        if self.fn in ["<=", "<", ">=", ">"]:
            actual_value = Decimal(list(value.values())[0])
            expected = Decimal(self.right["S"])
            if self.fn == "<=":
                return actual_value <= expected
            if self.fn == "<":
                return actual_value < expected
            if self.fn == ">=":
                return actual_value >= expected
            if self.fn == ">":
                return actual_value > expected
        if self.fn == "attribute_type" and value is not None:
            actual_value = list(value.keys())[0]
            return actual_value == self.right["S"]
        # Default - should we error instead if fn != '=='?
        return value == self.right

    def process_value(self, fn: Callable[[str], Dict[str, Any]]) -> None:
        self.right = fn(self.right)

    def get_filter_names(self) -> List[str]:
        return [".".join(self.left)]

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, WhereClause):
            return (
                self.fn == other.fn
                and self.left == other.left
                and self.right == other.right
            )
        return NotImplemented

    def __str__(self) -> str:
        return f"<{type(self)} {self.fn}({self.left}, {self.right})>"

    def __repr__(self) -> str:
        return str(self)


class WhereParser:
    def __init__(self, source_data: List[CaseInsensitiveDict]):
        self.source_data = source_data

    @classmethod
    def parse_where_clause(
        cls, where_clause: str, tokenizer: Optional[ClauseTokenizer] = None
    ) -> AbstractWhereClause:
        where_clause_parser = tokenizer or ClauseTokenizer(where_clause)
        current_clause: Optional[AbstractWhereClause] = None
        processing_function = False
        left = []
        fn: str = ""
        section: Optional[str] = "KEY"
        current_phrase = ""
        while True:
            c = where_clause_parser.next()
            if c is None:
                if section == "KEY" and current_phrase != "":
                    left.append(current_phrase)
                if section == "START_VALUE" and current_phrase != "":
                    current_clause = cls._determine_current_clause(
                        current_clause, left=left.copy(), fn=fn, right=current_phrase
                    )
                break
            if section == "KEY" and c == "(":
                if current_phrase == "":
                    # Process a subsection of the WHERE-clause
                    # .. and (sub-clause) and ...
                    next_clause = WhereParser.parse_where_clause(
                        where_clause="", tokenizer=where_clause_parser
                    )
                    if current_clause is None:
                        current_clause = next_clause
                    else:
                        current_clause.children.append(next_clause)
                    section = "END_VALUE"
                    where_clause_parser.skip_white_space()
                    continue
                else:
                    # Function
                    fn = current_phrase
                    current_phrase = ""
                    processing_function = True
                    continue
            if c == ")":
                if processing_function:
                    #               |
                    #               v
                    #  fn("key", val)
                    processing_function = False
                    continue
                else:
                    # Finished processing a subsection of the WHERE-clause
                    # .. and (sub-clause) and ...
                    return current_clause  # type: ignore[return-value]
            if c == ".":
                if section in ["KEY", "END_KEY"]:
                    if current_phrase != "":
                        left.append(current_phrase)
                    current_phrase = ""
                    section = "KEY"
                    continue
            if c in [","]:
                if section in ["END_KEY"]:
                    #          |
                    #          v
                    #  fn("key", val)
                    section = "START_VALUE"
                    where_clause_parser.skip_white_space()
                    continue
            if c in ['"', "'"]:
                if section == "KEY":
                    # collect everything between these quotes
                    left.append(where_clause_parser.next_until([c]))
                    # This could be the end of a key
                    #         |
                    #         v
                    #     "key" = val
                    #  fn("key", val)
                    #     "key".subkey = val
                    #
                    # Note that in the last example, the key isn't actually finished
                    # When we encounter a '.' next, will we return to processing a KEY
                    section = "END_KEY"
                    continue
                if section == "START_VALUE":
                    current_phrase = where_clause_parser.next_until([c])
                    section = "END_VALUE"
                    current_clause = cls._determine_current_clause(
                        current_clause, left=left.copy(), fn=fn, right=current_phrase
                    )
                    left.clear()
                    current_phrase = ""
                    where_clause_parser.skip_white_space()
                    continue
            if c in [" "] and section in ["KEY", "END_KEY"]:
                #         |
                #         v
                #   "key" = val
                #     key >= val
                if current_phrase != "":
                    left.append(current_phrase)
                current_phrase = ""
                fn = where_clause_parser.next_until([" "])
                if fn == "IS":
                    # Options:
                    #   IS MISSING
                    #   IS NOT MISSING
                    current_phrase = where_clause_parser.next_until([" "])
                    if current_phrase == "NOT":
                        current_phrase = (
                            f"{current_phrase} {where_clause_parser.next_until([' '])}"
                        )
                    current_clause = cls._determine_current_clause(
                        current_clause, left=left.copy(), fn=fn, right=current_phrase
                    )
                    left.clear()
                    current_phrase = ""
                    section = "END_VALUE"
                    continue
                where_clause_parser.skip_white_space()
                section = "START_VALUE"
                continue
            if c in [" "] and section == "START_VALUE":
                #            |
                #            v
                #   "key" = 0 AND ..
                current_clause = cls._determine_current_clause(
                    current_clause, left=left.copy(), fn=fn, right=current_phrase
                )
                left.clear()
                current_phrase = ""
                section = "END_VALUE"
                continue
            if c in [" "] and section == "END_VALUE":
                if current_phrase.upper() == "AND":
                    current_clause = WhereAndClause(current_clause)
                    current_phrase = ""
                    section = "KEY"
                    where_clause_parser.skip_white_space()
                elif current_phrase.upper() == "OR":
                    current_clause = WhereOrClause(current_clause)
                    current_phrase = ""
                    section = "KEY"
                    where_clause_parser.skip_white_space()
                continue
            if c in ["?"] and section == "START_VALUE":
                # Most values have to be surrounded by quotes
                # Question marks are parameters, and are valid values on their own
                current_clause = cls._determine_current_clause(
                    current_clause, left=left.copy(), fn=fn, right="?"
                )
                left.clear()
                section = "END_VALUE"  # Next step is to look for other key/value pairs
                continue
            if current_phrase == "" and section == "START_KEY":
                section = "KEY"
            if section in ["KEY", "VALUE", "START_VALUE", "END_VALUE"]:
                current_phrase += c
        return current_clause  # type: ignore[return-value]

    @classmethod
    def _determine_current_clause(
        cls,
        current_clause: Optional[AbstractWhereClause],
        left: List[str],
        fn: str,
        right: str,
    ) -> AbstractWhereClause:
        if current_clause is not None and isinstance(
            current_clause, (WhereAndClause, WhereOrClause)
        ):
            current_clause.children.append(
                WhereClause(fn=fn, left=left.copy(), right=right)
            )
            return current_clause
        else:
            return WhereClause(fn=fn, left=left.copy(), right=right)


class DynamoDBWhereParser(WhereParser):
    def parse(
        self, _where_clause: str, parameters: Optional[List[Dict[str, Any]]]
    ) -> List[CaseInsensitiveDict]:
        where_clause = WhereParser.parse_where_clause(_where_clause)

        def prep_value(val: str) -> Dict[str, Any]:
            # WHERE key = ?
            #     ? should be parametrized
            if val == "?" and parameters:
                return parameters.pop(0)
            # WHERE key = val
            #     'val' needs to be comparable with a DynamoDB document
            #     So we need to turn that into {'S': 'val'}
            else:
                return serializer.serialize(val)

        where_clause.process_value(prep_value)

        return [
            row
            for row in self.source_data
            if where_clause.apply(find_value_in_dynamodb_document, row)
        ]


class S3WhereParser(WhereParser):
    def parse(self, _where_clause: str) -> Any:
        # parameters argument is ignored - only relevant for DynamoDB
        where_clause = WhereParser.parse_where_clause(_where_clause)

        return [
            row
            for row in self.source_data
            if where_clause.apply(find_value_in_document, row)
        ]



from typing import Any, List, Tuple

from .clause_tokenizer import ClauseTokenizer
from .utils import serializer


class UpdateParser:
    def parse(
        self, query: str
    ) -> Tuple[str, List[Tuple[str, Any]], List[Tuple[str, Any]]]:
        tokenizer = ClauseTokenizer(query)

        section = "START"
        current_phrase = ""

        table_name = ""
        attrs_to_update: List[Tuple[str, Any]] = []
        attr_filters: List[Tuple[str, Any]] = []

        while True:
            c = tokenizer.next()
            if c is None:
                break

            if c == " ":
                if section == "START":
                    assert (
                        current_phrase.upper() == "UPDATE"
                    ), f"{current_phrase} should be UPDATE"
                    current_phrase = ""
                    section = "TABLE_NAME"
                    continue
                if section == "ACTION":
                    assert current_phrase.upper() in ["SET", "REMOVE"]
                    section = f"ACTION_{current_phrase.upper()}"
                    current_phrase = ""
                    continue
                if section == "SECTION_WHERE":
                    assert current_phrase.upper() == "WHERE"
                    section = "WHERE"
                    current_phrase = ""
                    continue
                if section == "WHERE_AND":
                    assert current_phrase.upper() == "AND"
                    section = "WHERE"
                    current_phrase = ""
                if section == "ACTION_REMOVE":
                    attr_name = current_phrase

                    attrs_to_update.append((attr_name, None))
                    current_phrase = ""

                    tokenizer.skip_white_space()
                    section = "SECTION_WHERE"
                if section == "TABLE_NAME":
                    table_name = current_phrase
                    current_phrase = ""
                    tokenizer.skip_white_space()
                    section = "ACTION"
                continue
            elif c in ["'", '"']:
                if section == "TABLE_NAME":
                    table_name = tokenizer.next_until([c])
                    tokenizer.skip_white_space()
                    section = "ACTION"
                if section == "ACTION_SET":
                    current_phrase = tokenizer.next_until([c])
                    tokenizer.skip_white_space()
                if section == "WHERE":
                    current_phrase = tokenizer.next_until([c])
                    tokenizer.skip_white_space()
                continue
            elif c == "=":
                if section == "ACTION_SET":
                    attr_name = current_phrase
                    tokenizer.skip_white_space()
                    if tokenizer.current() == "?":
                        attrs_to_update.append((attr_name, "?"))
                        tokenizer.next_until(["?"])
                    else:
                        quote_type = tokenizer.current()
                        assert quote_type in ["'", '"', "?"]

                        tokenizer.next()
                        attr_value = tokenizer.next_until([quote_type])
                        attrs_to_update.append(
                            (attr_name, serializer.serialize(attr_value))
                        )
                    current_phrase = ""

                    tokenizer.skip_white_space()
                    if tokenizer.current() == ",":
                        tokenizer.next_until([","])
                        # Another attr to update
                        pass
                    else:
                        section = "SECTION_WHERE"
                elif section == "WHERE":
                    attr_name = current_phrase
                    tokenizer.skip_white_space()
                    if tokenizer.current() == "?":
                        attr_filters.append((current_phrase, "?"))
                        tokenizer.next_until(["?"])
                    else:
                        quote_type = tokenizer.current()
                        assert quote_type in ["'", '"']

                        tokenizer.next()
                        attr_value = tokenizer.next_until([quote_type])
                        attr_filters.append(
                            (attr_name, serializer.serialize(attr_value))
                        )

                    tokenizer.skip_white_space()
                    current_phrase = ""
                    section = "WHERE_AND"
                continue
            else:
                current_phrase += c
                continue

        return table_name, attrs_to_update, attr_filters



from typing import Any, Dict

from .clause_tokenizer import ClauseTokenizer
from .json_parser import JsonParser
from .utils import CaseInsensitiveDict

from ..exceptions import ParserException


class FromParser:
    def __init__(self, from_clause: str):
        """
        Parse a FROM-clause in a PARTIQL query
        :param from_clause: a string of format `a AS b, x AS y` where `a` and `x` can contain commas
        :return: a dictionary of format `[b:a, y:x]`
        """
        clauses: Dict[str, Any] = dict()
        section = None  # NAME/AS/ALIAS
        current_phrase = ""
        name = None
        from_clause_parser = ClauseTokenizer(from_clause)
        while True:
            c = from_clause_parser.next()
            if c is None:
                break
            if c == "[":
                if section is None:
                    # Beginning of the FROM-clause - probably a document in its own right, instead of a name
                    current_phrase = "[" + from_clause_parser.next_until(["]"]) + "]"
                    section = "NAME"
                    continue
            if c == " ":
                if section == "AS" and current_phrase.upper() == "AS":
                    current_phrase = ""
                    section = "ALIAS"
                elif section == "NAME":
                    name = current_phrase
                    current_phrase = ""
                    section = "AS"
                elif section == "ALIAS":
                    current_phrase = ""
                    section = "NAME"
                continue
            if c == ",":
                if section == "NAME":
                    clauses[current_phrase] = current_phrase
                elif section == "ALIAS":
                    clauses[current_phrase] = name
                    section = "NAME"
                current_phrase = ""
                from_clause_parser.skip_white_space()
                # new phrase
                section = None
                continue

            if section is None:
                section = "NAME"

            current_phrase += c
        if section == "NAME":
            clauses[current_phrase] = current_phrase
        else:
            clauses[current_phrase] = name
        #
        # One FROM clause may point to the alias of another
        # {'s': 'sensors', 'r': 's.readings'} --> {'s': 'sensors', 'r': 'sensors.readings'}
        aliases = [(f"{key}.", f"{value}.") for key, value in clauses.items()]
        for key, value in clauses.items():
            for short, long in aliases:
                if value.startswith(short):
                    clauses[key] = value.replace(short, long)

        # {alias: full_name_of_table_or_file}
        self.clauses = clauses


class S3FromParser(FromParser):
    def get_source_data(self, documents: Dict[str, str]) -> Any:
        from_alias = list(self.clauses.keys())[0].lower()
        from_query = list(self.clauses.values())[0].lower()
        if "." in from_query:
            return self._get_nested_source_data(documents)

        key_has_asterix = from_query.endswith("[*]")
        from_query = from_query[0:-3] if key_has_asterix else from_query
        from_alias = from_alias[0:-3] if from_alias.endswith("[*]") else from_alias
        doc_is_list = documents[from_query].startswith("[") and documents[
            from_query
        ].endswith("]")

        source_data = JsonParser().parse(documents[from_query])

        if doc_is_list:
            return {"_1": source_data}
        elif from_alias:
            if isinstance(source_data, list):
                return [CaseInsensitiveDict({from_alias: doc}) for doc in source_data]
            else:
                return CaseInsensitiveDict({from_alias: source_data})
        else:
            return source_data

    def _get_nested_source_data(self, documents: Dict[str, Any]) -> Any:
        """
        Our FROM-clauses are nested, meaning we need to dig into the provided document to return the key that we need
           --> FROM s3object.name as name
        """
        root_doc = True
        source_data = documents
        iterate_over_docs = False
        entire_key = list(self.clauses.values())[0].lower().split(".")
        alias = list(self.clauses.keys())[0]
        if alias.endswith("[*]"):
            alias = alias[0:-3]
        key_so_far = []
        for key in entire_key:
            key_so_far.append(key)
            key_has_asterix = key.endswith("[*]") and key[0:-3] in source_data
            new_key = key[0:-3] if key_has_asterix else key
            if iterate_over_docs and isinstance(source_data, list):  # type: ignore[unreachable]
                # The previous key ended in [*]
                # Iterate over all docs in the result, and only return the requested source key
                if key_so_far == entire_key:  # type: ignore[unreachable]
                    # If we have an alias, we have to use that instead of the original name
                    source_data = [{alias: doc.get(new_key, {})} for doc in source_data]
                else:
                    source_data = [
                        doc.get_original(new_key) or CaseInsensitiveDict({})
                        for doc in source_data
                    ]
            else:
                # The previous key was a regular key
                # Assume that the result consists of a singular JSON document
                if new_key in source_data:
                    doc_is_list = source_data[new_key].startswith("[") and source_data[
                        new_key
                    ].endswith("]")
                    source_data = JsonParser().parse(source_data[new_key])
                    if root_doc and doc_is_list:
                        # AWS behaviour when the root-document is a list
                        source_data = {"_1": source_data}
                    elif key_so_far == entire_key:
                        if isinstance(source_data, list):  # type: ignore[unreachable]
                            source_data = [{alias: doc} for doc in source_data]  # type: ignore[unreachable]
                        else:
                            source_data = {alias: source_data}
                else:
                    source_data = {}

            iterate_over_docs = key_has_asterix
            root_doc = False

        return source_data


class DynamoDBFromParser(FromParser):
    def __init__(self, from_clause: str):
        super().__init__(from_clause)

        for alias, table_name in list(self.clauses.items()):
            if table_name[0].isnumeric():
                raise ParserException(
                    "ValidationException", "Aliasing is not supported"
                )

            if table_name[0] == '"' and table_name[-1] == '"':
                self.clauses[alias] = table_name[1:-1]

            if table_name[0] == "'" and table_name[-1] == "'":
                self.clauses[alias] = table_name[1:-1]



__version__ = "0.5.4"


from ._internal.parser import DynamoDBStatementParser, S3SelectParser  # noqa
from ._internal.json_parser import SelectEncoder  # noqa
from ._internal.csv_converter import csv_to_json, json_to_csv  # noqa
from ._internal.utils import MissingVariable, QueryMetadata  # noqa
